// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import superagent from 'superagent';
import ecDSAFormat from './ecDSAFormat';
import ElGamalKey from './elGamalKey';
import EcScalar from './ecScalar';
import TFastCi from './models/tfastCi';
import TFastRj from './models/tfastRj';
import EcScalarListSerializer from './ecScalarListSerializer';

export default class TSignClient {
    /**
     * 
     * @param {string} url 
     * @param {string} user 
     */
    constructor(url, user) {
        this.url = url + '/api';
        this.user = user;
        this._id = null;
    }

    async getId() {
        if (this._id != null)
            return Promise.resolve(this._id);

        const res = await superagent.get(`${this.url}/storage`);
        return this._id = BigInt(res.text);
    }

    async getSignKi(message) {
        var req = BigInt.isInstance(message)
            ? superagent.get(`${this.url}/sign/ki/${this.user}/${message}`)
            : superagent.post(`${this.url}/sign/ki/${this.user}`).send(JSON.stringify(message)).set('Content-Type', 'application/json');
        
        return (await req).text;
    }

    async getSignAlpha(ks) {
        const res = await superagent.post(`${this.url}/sign/alpha/${this.user}`).send(ks).set('Content-Type', 'application/json');
        return res.body;
    }

    async getSignDelta(beta, alphas) {
        const res = await superagent.post(`${this.url}/sign/delta/${this.user}/${encodeBase64Url(beta)}`).send(alphas).set('Content-Type', 'application/json');
        return res.body;
    }

    async getSignRSi(sigma, deltas) {
        const res = await superagent.post(`${this.url}/sign/rsi/${this.user}/${encodeBase64Url(sigma)}`).send(deltas).set('Content-Type', 'application/json');
        return RSiResponse.fromSignature(res.text);
    }

    /**
     * @param {EcScalar} pass
     * @param {number} threshold
     * @param {EcScalar[]} ids
     * @returns {Promise<TFastCi[]>}
     */
    async getFastCi(pass, threshold, ids) {
        const res = await superagent.get(`${this.url}/fastauth/ci/${this.user}/${pass}/${threshold}/${encodeBase64Url(EcScalarListSerializer.serializeStr(ids))}`);
        return res.body.map(data => TFastCi.from(data));
    }

    /**
     * @param {TFastCi[]} cs
     * @returns {Promise<TFastRj[]>}
     */
    async getFastRj(cs) {
        const res = await superagent.post(`${this.url}/fastauth/rj/${this.user}`).send(cs.map(ri => ri.toString())).set('Content-Type', 'application/json');
        return res.body.map(data => TFastRj.from(data));
    }

    /**
     * @param {TFastRj[]} rs
     * @param {ElGamalKey} pub
     * @returns {Promise<{priv: string, pub: ElGamalKey}>}
     */
    async getFastKey(rs, pub) {
        const res = await superagent.post(`${this.url}/fastauth/key/${this.user}/${encodeBase64Url(pub.toString())}`).send(rs.map(ri => ri.toString())).set('Content-Type', 'application/json');
        console.log(res);
        return { priv: res.body.private_key_frag, pub: ElGamalKey.from(res.body.public_key) };
    }

    async getAuthRi() {
        const res = await superagent.get(`${this.url}/auth/ri/${this.user}`);
        return res.text;
    }

    /**
     * @param {BigInt.BigInteger} pass
     * @param {string | object} rs
     * @returns {Promise<{private: string, public: string[]}>}
     */
    async getAuthAlpha(pass, rs) {
        const res = await superagent.post(`${this.url}/auth/alpha/${this.user}/${pass}`).send(rs).set('Content-Type', 'application/json');
        return res.body;
    }

    /**
     * @param {string} [beta]
     * @param {string[]} [alphas]
     * @param {ElGamalKey} pub
     * @returns {Promise<{priv: string, pub: ElGamalKey}>}
     */
    async getAuthSi(beta, pub, alphas) {
        try {
            const res = await superagent.post(`${this.url}/auth/si/${this.user}/${encodeBase64Url(beta)}/${encodeBase64Url(pub.toString())}`).send(alphas).set('Content-Type', 'application/json');
            var { private_key_frag, public_key }  = res.body.content.vendorFragment;
            
            return { priv: private_key_frag, pub: ElGamalKey.from(public_key)};
        } catch (error) { 
            if (OrkError.isInstance(error.response.body))
                throw new OrkError(error.response.body);
            
            throw error;
        }
    }

    /**
     * 
     * @param {EcScalar} password
     * @param {EcScalar} key
     */
    async SetAccount(password, key) {
        var prm = { password: password.toString(), key: ecDSAFormat.FormatPrivatePem(key) };
        await superagent.post(`${this.url}/storage/${this.user}`).send(prm).set('Content-Type', 'application/json');
    }

}

export class TSignGetClient extends TSignClient {
    constructor(url, user) {
        super(url, user);
    }

    async getSignAlpha(ks) {
        const res = await superagent.get(`${this.url}/getsign/alpha/${this.user}?${getQueryParam("ks", ks)}`);
        return res.body;
    }

    async getSignDelta(beta, alphas) {
        const res = await superagent.get(`${this.url}/getsign/delta/${this.user}/${encodeBase64Url(beta)}?${getQueryParam("alphas", alphas)}`);
        return res.body;
    }

    async getSignRSi(sigma, deltas) {
        const res = await superagent.get(`${this.url}/getsign/rsi/${this.user}/${encodeBase64Url(sigma)}?${getQueryParam("deltas", deltas)}`);
        return RSiResponse.fromSignature(res.text);
    }
};

function getQueryParam(name, data) {
    return data.map(itm => `${name}=${encodeBase64Url(itm)}`).join('&');
}

function encodeBase64Url(text) {
    return text.replace(/\=/g, "").replace(/\+/g, "-").replace(/\//g, "_");
}


class RSiResponse {
    /**
     * @param {BigInt.BigInteger} r
     * @param {BigInt.BigInteger} si
     * @param {BigInt.BigInteger} xi
     * @param {BigInt.BigInteger} yi
     */
    constructor(r, si, xi, yi) {
        this.r = r;
        this.si = si;
        this.xi = xi;
        this.yi = yi;
    }

    /**
     * @param {string} sign
     */
    static fromSignature(sign) {
        var array = Buffer.from(sign, 'base64');
        var r = BigInt.fromArray(Array.from(array.slice(0, 32)), 256, false);
        var si = BigInt.fromArray(Array.from(array.slice(32, 64)), 256, false);
        var xi = BigInt.fromArray(Array.from(array.slice(64, 96)), 256, false);
        var yi = BigInt.fromArray(Array.from(array.slice(96, 128)), 256, false);

        return new RSiResponse(r, si, xi, yi);
    }

    inspect() {
        return `<rsi r: ${this.r} si: ${this.si} xi: ${this.xi} yi: ${this.yi} >`;
    }

}

function decodeSignatureFromArray(signature) {
    var array = Buffer.from(signature, 'base64');
    var r = BigInt.fromArray(Array.from(array.slice(4).slice(0, array[3])), 256, false);
    var s = BigInt.fromArray(Array.from(array.slice(array[3] + 6)), 256, false);

    return [r, s];
}

class OrkError extends Error {
    /**
     * @typedef {{ success: boolean; error: string; }} Options
     * @param {string|Options} message
     */
    constructor(message) {
        super(typeof message === "string" ? message : message.error);
    }

    /**
     * @param {Options} body
     */
    static isInstance(body) {
        return body && body.success === false && body.error;
    }
}
